import logging

logger = logging.getLogger("maeluma-python-sdk")

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

logger.addHandler(ch)
logger.setLevel(logging.WARNING)
logger.propagate = False


def log_debug():
    logger.setLevel(logging.DEBUG)


def log_info():
    logger.setLevel(logging.INFO)


def log_warning():
    logger.setLevel(logging.WARNING)


def log_error():
    logger.setLevel(logging.ERROR)


def log_critical():
    logger.setLevel(logging.CRITICAL)


import html
from concurrent.futures import Future
from typing import Any, Callable, Dict, Iterator, Optional
from xmlrpc.client import Boolean


def _escape_html(text):
    return html.escape(str(text), quote=False)


def _df_html(
    df, style: Optional[Dict] = None, drop_all_na=True, dont_escape=("token_likelihoods",), **kwargs
):  # keep html in some columns
    formatters = {c: str if c in dont_escape else _escape_html for c in df.columns}
    if drop_all_na:  # do not show likelihood etc if all missing
        df = df.dropna(axis=1, how="all")
    if style:
        df = df.style.set_properties(**style)
    kwargs = dict(escape=False, formatters=formatters, **kwargs)
    return df.to_html(**kwargs)


class AsyncAttribute:
    """An attribute of an object that is lazily fetched.

    `async_request` is a Future object that is expected to resolve to an object that will be consumed by `getter`.
    `getter` is a function that recieves the result of `async_request` and processes it into the desired attribute.

    `getter` is only called once and its result is cached.
    """

    def __init__(self, async_request: Future, getter: Callable[..., Any]) -> None:
        self._request = async_request
        self._getter = getter
        self._resolved = False

    def __len__(self):
        return len(self.resolve())

    def __iter_(self) -> Iterator:
        return iter(self.resolve())

    def __repr__(self):
        return repr(self.resolve())

    def __str__(self):
        return str(self.resolve())

    def is_resolved(self) -> Boolean:
        return self._request.done()

    def resolve(self) -> Any:
        if "_result" in self.__dict__:
            return self._result

        self._result = self._getter(self._request.result())
        return self._result


class MaelumaObject:
    def __init__(self, id: str = None) -> None:
        self.id = id

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)
        if isinstance(attr, AsyncAttribute):
            return attr.resolve()
        else:
            return attr

    def __repr__(self) -> str:
        contents = ""
        exclude_list = ["iterator"]

        for k in self.__dict__.keys():
            if k not in exclude_list:
                contents += f"\t{k}: {self.__dict__[k]}\n"

        output = f"maeluma.{type(self).__name__} {{\n{contents}}}"
        return output

    def _repr_html_(self):  # rich html output for Jupyter
        try:
            return self.visualize()
        except (ImportError, AttributeError, NotImplementedError):  # no pandas or no visualize method
            return None  # ipython will use repr()


from typing import Any, Dict, Iterator, List, Optional

from maeluma.responses.base import MaelumaObject


class Embeddings(MaelumaObject):
    def __init__(
        self,
        embeddings: List[List[float]],
        compressed_embeddings: List[List[int]] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.embeddings = embeddings
        self.compressed_embeddings = compressed_embeddings
        self.meta = meta

    def __iter__(self) -> Iterator:
        return iter(self.embeddings)

    def __len__(self) -> int:
        return len(self.embeddings)


import json
from enum import Enum
from typing import Any, Dict, Generator, List, NamedTuple, Optional

import requests

from maeluma.responses.base import MaelumaObject


class Mode(str, Enum):
    CHAT = "chat"
    SEARCH_QUERY_GENERATION = "search_query_generation"
    AUGMENTED_GENERATION = "augmented_generation"


class Chat(MaelumaObject):
    """
    
    
    """
    def __init__(
        self,
        response_id: Optional[str],
        generation_id: Optional[str],
        message: Optional[str],
        text: Optional[str],
        conversation_id: Optional[str],
        meta: Optional[Dict[str, Any]] = None,
        prompt: Optional[str] = None,
        chatlog: Optional[List[Dict[str, str]]] = None,
        preamble: Optional[str] = None,
        token_count: Optional[Dict[str, int]] = None,
        client=None,
        queries: Optional[List[str]] = None,
        citations: Optional[List[Dict[str, str]]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.response_id = response_id
        self.generation_id = generation_id
        self.message = message
        self.text = text
        self.conversation_id = conversation_id
        self.prompt = prompt  # optional
        self.chatlog = chatlog  # optional
        self.preamble = preamble  # optional
        self.client = client
        self.token_count = token_count
        self.meta = meta
        self.queries = queries
        self.citations = citations

    @classmethod
    def from_dict(cls, response: Dict[str, Any], message: str, client) -> "Chat":
        return cls(
            id=response.get("response_id"),
            response_id=response.get("response_id"),
            generation_id=response.get("generation_id"),
            message=message,
            conversation_id=response["conversation_id"],
            text=response.get("text"),
            prompt=response.get("prompt"),  # optional
            chatlog=response.get("chatlog"),  # optional
            preamble=response.get("preamble"),  # option
            client=client,
            token_count=response.get("token_count"),
            meta=response.get("meta"),
            queries=response.get("queries"),
            citations=response.get("citations"),
        )

    def respond(self, response: str, max_tokens: int = None) -> "Chat":
        return self.client.chat(
            message=response,
            conversation_id=self.conversation_id,
            return_chatlog=self.chatlog is not None,
            return_prompt=self.prompt is not None,
            return_preamble=self.preamble is not None,
            max_tokens=max_tokens,
        )


class AsyncChat(Chat):
    async def respond(self, response: str, max_tokens: int = None) -> "AsyncChat":
        return await self.client.chat(
            message=response,
            conversation_id=self.conversation_id,
            return_chatlog=self.chatlog is not None,
            return_prompt=self.prompt is not None,
            return_preamble=self.preamble is not None,
            max_tokens=max_tokens,
        )


StreamingText = NamedTuple("StreamingText", [("index", Optional[int]), ("text", str), ("is_finished", bool)])


class StreamingChat(MaelumaObject):
    def __init__(self, response):
        self.response = response
        self.texts = []
        self.response_id = None
        self.conversation_id = None
        self.preamble = None
        self.prompt = None
        self.chatlog = None
        self.finish_reason = None

    def _make_response_item(self, index, line) -> Any:
        streaming_item = json.loads(line)
        is_finished = streaming_item.get("is_finished")
        text = streaming_item.get("text")

        if not is_finished:
            return StreamingText(text=text, is_finished=is_finished, index=index)

        response = streaming_item.get("response")

        if response is None:
            return None

        self.response_id = response.get("response_id")
        self.conversation_id = response.get("conversation_id")
        self.preamble = response.get("preamble")
        self.prompt = response.get("prompt")
        self.chatlog = response.get("chatlog")
        self.finish_reason = streaming_item.get("finish_reason")
        self.texts = [response.get("text")]
        return None

    def __iter__(self) -> Generator[StreamingText, None, None]:
        if not isinstance(self.response, requests.Response):
            raise ValueError("For AsyncClient, use `async for` to iterate through the `StreamingChat`")

        for index, line in enumerate(self.response.iter_lines()):
            item = self._make_response_item(index, line)
            if item is not None:
                yield item

    async def __aiter__(self) -> Generator[StreamingText, None, None]:
        index = 0
        async for line in self.response.content:
            item = self._make_response_item(index, line)
            index += 1
            if item is not None:
                yield item


import json
from collections import UserList
from typing import Any, Dict, Generator, List, NamedTuple, Optional

import requests

from maeluma.responses.base import MaelumaObject, _df_html

TokenLikelihood = NamedTuple("TokenLikelihood", [("token", str), ("likelihood", float)])

TOKEN_COLORS = [
    (-2, "#FFECE2"),
    (-4, "#FFD6BC"),
    (-6, "#FFC59A"),
    (-8, "#FFB471"),
    (-10, "#FFA745"),
    (-12, "#FE9F00"),
    (-1e9, "#E18C00"),
]


class ImageGeneration(MaelumaObject, str):
    def __new__(cls, text: str, *_, **__):
        return str.__new__(cls, text)

    def __init__(
        self,
        prompt: str,
        width: int,
        height: int,
        negative_prompt: str,
        url: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.prompt = prompt
        self.width = width
        self.height = height
        self.negative_prompt = negative_prompt
        self.url = url

    @classmethod
    def from_response(cls, response, prompt=None, **kwargs):
        image_metadata = response.get("meta")
        
        return cls(
            prompt=response.get("prompt"),
            width=image_metadata.get("width", 512),
            height=image_metadata.get('height', 512),
            negative_prompt=response.get('negative_prompt', ''),
            id=image_metadata.get("file_prefix", response.get("output")[0]),
            url=response.get("output")[0],
            **kwargs,
        )


class ImageGenerations(UserList, MaelumaObject):
    def __init__(self, generations,  meta: Optional[Dict[str, Any]] = None) -> None:
        super().__init__(generations)
        self.meta = meta

    @classmethod
    def from_dict(cls, response: Dict[str, Any]) -> List[ImageGeneration]:
        generations: List[ImageGeneration] = []
        for gen in response["generations"]:
            generations.append(
                ImageGeneration(
                    gen["text"],
                    likelihood,
                    token_likelihoods,
                    prompt=response.get("prompt"),
                    id=gen["id"],
                    finish_reason=gen.get("finish_reason"),
                )
            )

        return cls(generations, response.get("meta"))

    @property
    def generations(self) -> List[ImageGeneration]:  # backward compatibility
        return self.data

    @property
    def prompt(self) -> str:
        """Returns the prompt used as input"""
        return self[0].prompt  # should all be the same


# ("likelihood", Optional[float])]) not supported
StreamingText = NamedTuple("StreamingText", [("index", Optional[int]), ("text", str), ("is_finished", bool)])


class StreamingImageGenerations(MaelumaObject):
    def __init__(self, response):
        self.response = response
        self.id = None
        self.generations = None
        self.finish_reason = None
        self.texts = []

    def _make_response_item(self, line) -> Optional[StreamingText]:
        streaming_item = json.loads(line)
        is_finished = streaming_item.get("is_finished")

        if not is_finished:
            index = streaming_item.get("index", 0)
            text = streaming_item.get("text")
            while len(self.texts) <= index:
                self.texts.append("")
            if text is None:
                return None
            self.texts[index] += text
            return StreamingText(text=text, is_finished=is_finished, index=index)

        self.finish_reason = streaming_item.get("finish_reason")
        generation_response = streaming_item.get("response")

        if generation_response is None:
            return None

        self.id = generation_response.get("id")
        # likelihoods not supported in streaming currently
        self.generations = ImageGeneration.from_dict(generation_response)
        return None

    def __iter__(self) -> Generator[StreamingText, None, None]:
        if not isinstance(self.response, requests.Response):
            raise ValueError("For AsyncClient, use `async for` to iterate through the `StreamingGenerations`")

        for line in self.response.iter_lines():
            item = self._make_response_item(line)
            if item is not None:
                yield item

    async def __aiter__(self) -> Generator[StreamingText, None, None]:
        async for line in self.response.content:
            item = self._make_response_item(line)
            if item is not None:
                yield item


from enum import Enum


class LengthEnum(str, Enum):
    LONG = 'long',
    MEDIUM = 'medium',
    SHORT = 'short',
    AUTO = 'auto',

class VolumeEnum(str, Enum):
    """
      A General Enum for the for high, medium, low and auto
    """
    HIGH = 'HIGH',
    MEDIUM = 'medium',
    LOW = 'low',
    AUTO = 'auto',


class TextFormat(str, Enum):
    PARAGRAPHS = 'paragraph',
    BULLETS = 'bullets',
    AUTO = 'auto',


from typing import Dict


class MaelumaError(Exception):
    """Base Exception class, returned when nothing more specific applies"""

    def __init__(self, message=None) -> None:
        super(MaelumaError, self).__init__(message)

        self.message = message

    def __str__(self) -> str:
        msg = self.message or "<empty message>"
        return msg

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(message={str(self)})"


class MaelumaAPIError(MaelumaError):
    """Returned when the API responds with an error message"""

    def __init__(self, message: str = None, http_status: int = None, headers: Dict = None):
        super().__init__(message)
        self.http_status = http_status
        self.headers = headers or {}

    @classmethod
    def from_response(cls, response, message=None):
        return cls(message=message or response.text, http_status=response.status, headers=response.headers)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(message={str(self)}, http_status={self.http_status})"


class MaelumaConnectionError(MaelumaError):
    """Returned when the SDK can not reach the API server for any reason"""


import json as jsonlib
import os
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict
from functools import partial
from typing import Any, Dict, Iterable, List, Optional, Union
from io import BytesIO, BufferedReader
from pathlib import Path  
try:
    from typing import Literal, TypedDict
except ImportError:
    from typing_extensions import Literal, TypedDict

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

import maeluma
from maeluma.enums import VolumeEnum, TextFormat, LengthEnum
from maeluma.custom_model_dataset import CustomModelDataset
from maeluma.error import MaelumaAPIError, MaelumaConnectionError, MaelumaError
from maeluma.logging import logger
from maeluma.responses import (
    Detokenization,
    TextGenerations,
    StreamingTextGenerations,
    Tokens,
    ImageGenerations,
    UpscaledImage,
    SummarizeResponse,
)
from maeluma.responses.chat import Chat, Mode, StreamingChat
#from maeluma.responses.classify import Example as ClassifyExample
#from maeluma.responses.classify import LabelPrediction
#from maeluma.responses.cluster import ClusterJobResult, CreateClusterJobResponse
#from maeluma.responses.custom_model import (
#    CUSTOM_MODEL_PRODUCT_MAPPING,
#    CUSTOM_MODEL_STATUS,
#    CUSTOM_MODEL_TYPE,
#    INTERNAL_CUSTOM_MODEL_TYPE,
#    CustomModel,
#    HyperParametersInput,
#
# )
# from maeluma.responses.detectlang import DetectLanguageResponse, Language
from maeluma.responses.embeddings import Embeddings

from maeluma.utils import is_api_key_valid, threadpool_map, wait_for_job


class Client:
    """Cohere Client

    Args:
        api_key (str): Your API key.
        num_workers (int): Maximal number of threads for parallelized calls.
        request_dict (dict): Additional parameters for calls with the requests library. Currently ignored in AsyncClient
        check_api_key (bool): Whether to check the api key for validity on initialization.
        client_name (str): A string to identify your application for internal analytics purposes.
        max_retries (int): maximal number of retries for requests.
        timeout (int): request timeout in seconds.
    """

    def __init__(
        self,
        api_key: str = None,
        num_workers: int = 64,
        request_dict: dict = {},
        check_api_key: bool = True,
        client_name: Optional[str] = None,
        max_retries: int = 3,
        timeout: int = 120,
    ) -> None:
        self.api_key = api_key or os.getenv("MAE_API_KEY")
        self.api_url = os.getenv("MAE_API_URL", maeluma.MAE_API_URL)
        self.batch_size = maeluma.MAELUMA_EMBED_BATCH_SIZE
        self._executor = ThreadPoolExecutor(num_workers)
        self.num_workers = num_workers
        self.request_dict = request_dict
        self.request_source = "python-sdk-" + maeluma.SDK_VERSION
        self.max_retries = max_retries
        self.timeout = timeout
        self.api_version = f"v{maeluma.API_VERSION}"
        if client_name:
            self.request_source += ":" + client_name

        if check_api_key:
            self.check_api_key()

    def check_api_key(self) -> Dict[str, bool]:
        """
        Checks the api key, which happens automatically during Client initialization, but not in AsyncClient.
        check_api_key raises an exception when the key is invalid, but the return value for valid keys is kept for
        backwards compatibility.
        """
        return {"valid": is_api_key_valid(self.api_key)}

    def batch_generate(
        self, prompts: List[str], return_exceptions=False, **kwargs
    ) -> List[Union[TextGenerations, Exception]]:
        """A batched version of generate with multiple prompts.

        Args:
            prompts: list of prompts
            return_exceptions (bool): Return exceptions as list items rather than raise them. Ensures your entire batch is not lost on one of the items failing.
            kwargs: other arguments to `generate`
        """
        return threadpool_map(
            self.generate,
            [dict(prompt=prompt, **kwargs) for prompt in prompts],
            num_workers=self.num_workers,
            return_exceptions=return_exceptions,
        )

    def generate(
        self,
        prompt: Optional[str] = None,
        prompt_vars: object = {},
        model: Optional[str] = None,
        preset: Optional[str] = None,
        num_generations: Optional[int] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        k: Optional[int] = None,
        p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        end_sequences: Optional[List[str]] = None,
        stop_sequences: Optional[List[str]] = None,
        truncate: Optional[str] = None,
        stream: bool = False,
        logit_bias: Dict[int, float] = {},
    ) -> Union[TextGenerations, StreamingTextGenerations]:
        """Generate endpoint.
        See https://docs.maeluma.ai/reference/generate for advanced arguments

        Args:
            prompt (str): Represents the prompt or text to be completed. Trailing whitespaces will be trimmed.
            model (str): (Optional) The model ID to use for generating the next reply.
            preset (str): (Optional) The ID of a custom playground preset.
            num_generations (int): (Optional) The number of generations that will be returned, defaults to 1.
            max_tokens (int): (Optional) The number of tokens to predict per generation, defaults to 20.
            temperature (float): (Optional) The degree of randomness in generations from 0.0 to 5.0, lower is less random.
            truncate (str): (Optional) One of NONE|START|END, defaults to END. How the API handles text longer than the maximum token length.\
            stream (bool): Return streaming tokens.
        Returns:
            if stream=False: a Generations object

        Examples:
            A simple generate message:
                >>> res = co.generate(prompt="Hey! How are you doing today?")
                >>> print(res.text)
            Streaming generate:
                >>> res = co.generate(
                >>>     prompt="Hey! How are you doing today?",
                >>>     stream=True)
                >>> for token in res:
                >>>     print(token)
        """
        json_body = {
            "model": model,
            "prompt": prompt,
            "prompt_vars": prompt_vars,
            "preset": preset,
            "num_generations": num_generations,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "k": k,
            "p": p,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "end_sequences": end_sequences,
            "stop_sequences": stop_sequences,
            "truncate": truncate,
            "logit_bias": logit_bias,
            "stream": stream,
        }
        response = self._request(maeluma.GENERATE_URL, json=json_body, stream=stream)
        return TextGenerations.from_dict(response=response)


    def generate_image(
            self,
            prompt:str,
            num_generations: Optional[int] = 1,
            negative_prompt: Optional[str] = None,
            model: Optional[str] = 'general',
            width: Optional[int] = 512,
            height: Optional[int] = 512,
            samples: Optional[int] = 1,
            enhance_prompt=False,
            num_inference_steps: Optional[int] = 21,
            seed: Optional[int] = None,
            guidance_scale: Optional[float] = 7.5,
            safety_checker: Optional[bool] = False,
            multi_lingual: Optional[bool] = True,
            panorama: Optional[bool] = False,
            self_attention: Optional[bool] = None,
            upscale: Optional[bool] = None,
            embeddings_model: Optional[bool] = None,
            webhook: Optional[bool] = None,
            idempotency_key=None, 
            stream=False):
        json_body = {
          "key": idempotency_key,
          "prompt": prompt,
          'num_generations': num_generations,
          "negative_prompt": negative_prompt,
          'enhance_prompt': enhance_prompt,
          "width": width,
          "height": height,
          "samples": samples,
          'model': model, # 'general' or 'faces
          "num_inference_steps": num_inference_steps,
          "seed": seed,
          "guidance_scale": guidance_scale,
          "safety_checker": safety_checker,
          "multi_lingual": multi_lingual,
          "panorama": panorama,
          "self_attention": self_attention,
          "upscale": upscale,
          "embeddings_model": None,
          "webhook": None,
          "track_id": None
        }
        response = self._request(maeluma.GENERATE_IMAGE_URL, json=json_body, stream=stream)
        return response
    
    def batch_generate_image(
        self, prompts: List[str], return_exceptions=False, **kwargs
    ) -> List[Union[ImageGenerations, Exception]]:
        """A batched version of generate with multiple prompts.

        Args:
            prompts: list of prompts
            return_exceptions (bool): Return exceptions as list items rather than raise them. Ensures your entire batch is not lost on one of the items failing.
            kwargs: other arguments to `generate`
        """
        return threadpool_map(
            self.generate_image,
            [dict(prompt=prompt, **kwargs) for prompt in prompts],
            num_workers=self.num_workers,
            return_exceptions=return_exceptions,
        )

    
    def chat(
        self,
        message: Optional[str] = None,
        query: Optional[str] = None,
        conversation_id: Optional[str] = "",
        model: Optional[str] = None,
        return_chatlog: Optional[bool] = False,
        return_prompt: Optional[bool] = False,
        return_preamble: Optional[bool] = False,
        chat_history: Optional[List[Dict[str, str]]] = None,
        preamble_override: Optional[str] = None,
        user_name: Optional[str] = None,
        temperature: Optional[float] = 0.8,
        max_tokens: Optional[int] = None,
        stream: Optional[bool] = False,
        p: Optional[float] = None,
        k: Optional[float] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        mode: Optional[Mode] = None,
        documents: Optional[List[Dict[str, str]]] = None,
    ) -> Union[Chat, StreamingChat]:
        """Returns a Chat object with the query reply.

        Args:
            message (str): The message to send to the chatbot.
            conversation_id (str): (Optional) The conversation id to continue the conversation.
            model (str): (Optional) The model to use for generating the next reply.
            return_chatlog (bool): (Optional) Whether to return the chatlog.
            return_prompt (bool): (Optional) Whether to return the prompt.
            return_preamble (bool): (Optional) Whether to return the preamble.
            chat_history (List[Dict[str, str]]): (Optional) A list of entries used to construct the conversation. If provided, these messages will be used to build the prompt and the conversation_id will be ignored so no data will be stored to maintain state.
            preamble_override (str): (Optional) A string to override the preamble.
            user_name (str): (Optional) A string to override the username.
            temperature (float): (Optional) The temperature to use for the next reply. The higher the temperature, the more random the reply.
            max_tokens (int): (Optional) The max tokens generated for the next reply.
            stream (bool): Return streaming tokens.
            p (float): (Optional) The nucleus sampling probability.
            k (float): (Optional) The top-k sampling probability.
            logit_bias (Dict[int, float]): (Optional) A dictionary of logit bias values to use for the next reply.
            mode Mode: (Optional) This property determines which functionality of retrieval augmented generation to use.
                                    chat mode doesn't use any retrieval augmented generation functionality.
                                    search_query_generation uses the provided query to produce search terms that you can use to search for documents.
                                    augmented_generation uses the provided documents and query to produce citations
            document Document: (Optional) The documents to use in augmented_generation mode. Shape: ("title", str), ("snippet", str), ("url", str)
        Returns:
            a Chat object if stream=False, or a StreamingChat object if stream=True

        Examples:
            A simple chat message:
                >>> res = co.chat(message="Hey! How are you doing today?")
                >>> print(res.text)
                >>> print(res.conversation_id)
            Continuing a session using a specific model:
                >>> res = co.chat(
                >>>     message="Hey! How are you doing today?",
                >>>     conversation_id="1234",
                >>>     model="command",
                >>>     return_chatlog=True)
                >>> print(res.text)
                >>> print(res.chatlog)
            Streaming chat:
                >>> res = co.chat(
                >>>     message="Hey! How are you doing today?",
                >>>     stream=True)
                >>> for token in res:
                >>>     print(token)
            Stateless chat with chat history:
                >>> res = co.chat(
                >>>     message="Tell me a joke!",
                >>>     chat_history=[
                >>>         {'user_name': 'User', message': 'Hey! How are you doing today?'},
                >>>         {'user_name': 'Bot', message': 'I am doing great! How can I help you?'},
                >>>     ],
                >>>     return_prompt=True)
                >>> print(res.text)
                >>> print(res.prompt)
            Augmented generation example:
                >>> res = co.chat(query="What are the tallest penguins?",
                                  mode="augmented_generation",
                                  documents = [{"title":"Tall penguins", "snippet":"Emperor penguins are the tallest", "url":"http://example.com/foo"}])
                >>> print(res.text)
                >>> print(res.citations)
        """
        if chat_history is not None:
            should_warn = True
            for entry in chat_history:
                if "text" in entry:
                    entry["message"] = entry["text"]

                if "text" in entry and should_warn:
                    logger.warning(
                        "The 'text' parameter is deprecated and will be removed in a future version of this function. "
                        + "Use 'message' instead.",
                    )
                    should_warn = False

        if query is not None:
            logger.warning(
                "The chat_history 'text' key is deprecated and will be removed in a future version of this function. "
                + "Use 'message' instead.",
            )
            message = query

        json_body = {
            "message": message,
            "conversation_id": conversation_id,
            "model": model,
            "return_chatlog": return_chatlog,
            "return_prompt": return_prompt,
            "return_preamble": return_preamble,
            "chat_history": chat_history,
            "preamble_override": preamble_override,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": stream,
            "user_name": user_name,
            "p": p,
            "k": k,
            "logit_bias": logit_bias,
            "mode": mode,
            "documents": documents,
        }
        response = self._request(maeluma.CHAT_URL, json=json_body, stream=stream)

        if stream:
            return StreamingChat(response)
        else:
            return Chat.from_dict(response, message=message, client=self)


    def upscale_image(
            self,
            url: Optional[str] = None,
            image: Optional[Union[str, BytesIO]] = None,
            filetype: Optional[str] = "png",
    ):
      """
        Returns the url of the upscaled image.
        Args:
            url (str): The url of the image to upscale.
            image (Union[str, bytes, BytesIO]): (Optional) The image or image file path to upscale. If provided, the url will be ignored.
      """
      headers = {"Content-Disposition": "attachment; filename={url}"}
      files = {}
      if image is None and url is None:
          raise MaelumaError("You must provide either a url or an image.")
      if image is not None:
          files={"image": image}
          if isinstance(image, str):
              filename = Path(image).name
              image = self._get_image(image)
              headers = {"Content-Disposition": "attachment; filename={filename}".format(filename=filename)}
          else:
              filename = "image." + filetype
              headers = {"Content-Disposition": "attachment; filename={filename}".format(filename=filename)}
      else:
          image = url
      response = self._request(maeluma.UPSCALE_IMAGE_URL, json={"url": url}, files=files, headers=headers)
      return UpscaledImage(**response)
      

    def summarize(self, 
                  text: str,
                  length: Optional[LengthEnum] = LengthEnum.AUTO,
                  format: Optional[TextFormat] = TextFormat.AUTO,
                  model: Optional[str] = 'general',
                  extractiveness: Optional[VolumeEnum] = VolumeEnum.AUTO,
                  temperature: Optional[float] = 0.75,
                  additional_command: Optional[str] = '',
                ) -> SummarizeResponse:
        """Returns a generated summary of the specified length for the provided text.

        Args:
            text (str): Text to summarize.
            model (str): (Optional) ID of the model.
            length (str): (Optional) One of {"short", "medium", "long"}, defaults to "medium". \
                Controls the length of the summary.
            format (str): (Optional) One of {"paragraph", "bullets"}, defaults to "paragraph". \
                Controls the format of the summary.
            extractiveness (str) One of {"high", "medium", "low"}, defaults to "high". \
                Controls how close to the original text the summary is. "High" extractiveness \
                summaries will lean towards reusing sentences verbatim, while "low" extractiveness \
                summaries will tend to paraphrase more.
            temperature (float): Ranges from 0 to 5. Controls the randomness of the output. \
                Lower values tend to generate more “predictable” output, while higher values \
                tend to generate more “creative” output. The sweet spot is typically between 0 and 1.
            additional_command (str): (Optional) Modifier for the underlying prompt, must \
                complete the sentence "Generate a summary _".

        Examples:
            Summarize a text:
                >>> res = ml.summarize(text="Stock market report for today...")
                >>> print(res.summary)

            Summarize a text with a specific model and prompt:
                >>> res = ml.summarize(
                >>>     text="Stock market report for today...",
                >>>     model="summarize-xlarge",
                >>>     length="long",
                >>>     format="bullets",
                >>>     temperature=0.3,
                >>>     additional_command="focusing on the highest performing stocks")
                >>> print(res.summary)
        """
        json_body = {
            "text": text,
            "length": length,
            "format": format,
            "model": model,
            "extractiveness": extractiveness,
            "temperature": temperature,
            "additional_command": additional_command,
        }
        json_body = {k: v for k, v in json_body.items() if v is not None}
        response = self._request(maeluma.SUMMARIZE_URL, json=json_body) 
        return SummarizeResponse(id=response["id"], summary=response["summary"], meta=response["meta"])


    def embed(
        self,
        texts: List[str],
        model: Optional[str] = None,
        truncate: Optional[str] = None,
        compress: Optional[bool] = False,
        compression_codebook: Optional[str] = "default",
    ) -> Embeddings:
        """Returns an Embeddings object for the provided texts. Visit https://maeluma.ai/embed to learn about embeddings.

        Args:
            text (List[str]): A list of strings to embed.
            model (str): (Optional) The model ID to use for embedding the text.
            truncate (str): (Optional) One of NONE|START|END, defaults to END. How the API handles text longer than the maximum token length.
            compress (bool): (Optional) Whether to compress the embeddings. When True, the compressed_embeddings will be returned as integers in the range [0, 255].
            compression_codebook (str): (Optional) The compression codebook to use for compressed embeddings. Defaults to "default".
        """
        responses = {
            "embeddings": [],
            "compressed_embeddings": [],
        }
        json_bodys = []

        for i in range(0, len(texts), self.batch_size):
            texts_batch = texts[i : i + self.batch_size]
            json_bodys.append(
                {
                    "model": model,
                    "texts": texts_batch,
                    "truncate": truncate,
                    "compress": compress,
                    "compression_codebook": compression_codebook,
                }
            )

        meta = None
        for result in self._executor.map(lambda json_body: self._request(maeluma.EMBED_URL, json=json_body), json_bodys):
            responses["embeddings"].extend(result["embeddings"])
            responses["compressed_embeddings"].extend(result.get("compressed_embeddings", []))
            meta = result["meta"] if not meta else meta

        return Embeddings(
            embeddings=responses["embeddings"],
            compressed_embeddings=responses["compressed_embeddings"],
            meta=meta,
        )


    def batch_tokenize(self, texts: List[str], return_exceptions=False, **kwargs) -> List[Union[Tokens, Exception]]:
        """A batched version of tokenize.

        Args:
            texts: list of texts
            return_exceptions (bool): Return exceptions as list items rather than raise them. Ensures your entire batch is not lost on one of the items failing.
            kwargs: other arguments to `tokenize`
        """
        return threadpool_map(
            self.tokenize,
            [dict(text=text, **kwargs) for text in texts],
            num_workers=self.num_workers,
            return_exceptions=return_exceptions,
        )

    def tokenize(self, text: str, model: Optional[str] = None) -> Tokens:
        """Returns a Tokens object of the provided text, see https://docs.maeluma.ai/reference/tokenize for advanced usage.

        Args:
            text (str): Text to summarize.
            model (str): An optional model name that will ensure that the tokenization uses the tokenizer used by that model, which can be critical for counting tokens properly.
        """
        json_body = {"text": text, "model": model}
        res = self._request(maeluma.TOKENIZE_URL, json=json_body)
        return Tokens(tokens=res["tokens"], token_strings=res["token_strings"], meta=res.get("meta"))

    def batch_detokenize(
        self, list_of_tokens: List[List[int]], return_exceptions=False, **kwargs
    ) -> List[Union[Detokenization, Exception]]:
        """A batched version of detokenize.

        Args:
            list_of_tokens: list of list of tokens
            return_exceptions (bool): Return exceptions as list items rather than raise them. Ensures your entire batch is not lost on one of the items failing.
            kwargs: other arguments to `detokenize`
        """
        return threadpool_map(
            self.detokenize,
            [dict(tokens=tokens, **kwargs) for tokens in list_of_tokens],
            num_workers=self.num_workers,
            return_exceptions=return_exceptions,
        )

    def detokenize(self, tokens: List[int], model: Optional[str] = None) -> Detokenization:
        """Returns a Detokenization object of the provided tokens, see https://docs.maeluma.ai/reference/detokenize for advanced usage.

        Args:
            tokens (List[int]): A list of tokens to convert to strings
            model (str): An optional model name. This will ensure that the detokenization is done by the tokenizer used by that model.
        """
        json_body = {"tokens": tokens, "model": model}
        res = self._request(maeluma.DETOKENIZE_URL, json=json_body)
        return Detokenization(text=res["text"], meta=res.get("meta"))


    def _check_response(self, json_response: Dict, headers: Dict, status_code: int):
        if "X-API-Warning" in headers:
            logger.warning(headers["X-API-Warning"])
        if "message" in json_response:  # has errors
            raise MaelumaAPIError(
                message=json_response["message"],
                http_status=status_code,
                headers=headers,
            )
        if 400 <= status_code < 500:
            raise MaelumaAPIError(
                message=f"Unexpected client error (status {status_code}): {json_response}",
                http_status=status_code,
                headers=headers,
            )
        if status_code >= 500:
            raise MaelumaError(message=f"Unexpected server error (status {status_code}): {json_response}")
    

    def _get_image(self, file_path: str) -> BufferedReader:
        """Returns a file object of the image at the provided file path."""
        return open(file_path, "rb")
    

    def _request(self, endpoint, json=None, method="POST", stream=False, files: dict = {}, headers={}) -> Any:
        headers = {
            "Authorization": "BEARER {}".format(self.api_key),
            "Content-Type": "application/json",
            "Request-Source": self.request_source,
            **headers,
        }

        url = f"{self.api_url}/{self.api_version}/{endpoint}"
        with requests.Session() as session:
            retries = Retry(
                total=self.max_retries,
                backoff_factor=0.5,
                allowed_methods=["POST", "GET"],
                status_forcelist=maeluma.RETRY_STATUS_CODES,
                raise_on_status=False,
            )
            session.mount("https://", HTTPAdapter(max_retries=retries))
            session.mount("http://", HTTPAdapter(max_retries=retries))

            if stream:
                return session.request(method, url, headers=headers, json=json, files=files, **self.request_dict, stream=True)

            try:
                response = session.request(
                    method, url, headers=headers, json=json, timeout=self.timeout, files=files, **self.request_dict
                )
            except requests.exceptions.ConnectionError as e:
                raise MaelumaConnectionError(str(e)) from e
            except requests.exceptions.RequestException as e:
                raise MaelumaError(f"Unexpected exception ({e.__class__.__name__}): {e}") from e

            try:
                json_response = response.json()
            except jsonlib.decoder.JSONDecodeError:  # MaelumaAPIError will capture status
                raise MaelumaAPIError.from_response(response, message=f"Failed to decode json body: {response.text}")

            self._check_response(json_response, response.headers, response.status_code)
        return json_response


import asyncio
import json as jsonlib
import os
import posixpath
import time
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime, timezone
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

try:
    from typing import Literal, TypedDict
except ImportError:
    from typing_extensions import Literal, TypedDict

import aiohttp
import backoff

import maeluma
from maeluma.client import Client
from maeluma.custom_model_dataset import CustomModelDataset
from maeluma.error import MaelumaAPIError, MaelumaConnectionError, MaelumaError
from maeluma.logging import logger
from maeluma.responses import (
    Detokenization,
    Embeddings,
    TextGenerations,
    StreamingTextGenerations,
    Tokens,
)
from maeluma.responses.chat import AsyncChat, Mode, StreamingChat
from maeluma.utils import async_wait_for_job, is_api_key_valid, np_json_dumps

JSON = Union[Dict, List]


class AsyncClient(Client):
    """AsyncClient

    This client provides an asyncio/aiohttp interface.
    Using this client is recommended when you are making highly parallel request,
    or when calling the Cohere API from a server such as FastAPI."""

    def __init__(
        self,
        api_key: str = None,
        num_workers: int = 16,
        request_dict: dict = {},
        check_api_key: bool = True,
        client_name: Optional[str] = None,
        max_retries: int = 3,
        timeout=120,
    ) -> None:
        self.api_key = api_key or os.getenv("CO_API_KEY")
        self.api_url = os.getenv("CO_API_URL", maeluma.MAE_API_URL)
        self.batch_size = maeluma.MAELUMA_EMBED_BATCH_SIZE
        self.num_workers = num_workers
        self.request_dict = request_dict
        self.request_source = "python-sdk-" + maeluma.SDK_VERSION
        self.max_retries = max_retries
        if client_name:
            self.request_source += ":" + client_name
        self.api_version = f"v{maeluma.API_VERSION}"
        self._check_api_key_on_enter = check_api_key
        self._backend = AIOHTTPBackend(logger, num_workers, max_retries, timeout)

    async def _request(self, endpoint, json=None, method="POST", full_url=None, stream=False) -> JSON:
        headers = {
            "Authorization": f"BEARER {self.api_key}",
            "Request-Source": self.request_source,
        }
        if endpoint is None and full_url is not None:  # api key
            url = full_url
        else:
            url = posixpath.join(self.api_url, self.api_version, endpoint)

        response = await self._backend.request(url, json, method, headers, stream=stream)
        if stream:
            return response

        try:
            json_response = await response.json()
        #   `MaelumaAPIError.from_response()` will capture the http status code
        except jsonlib.decoder.JSONDecodeError:
            raise MaelumaAPIError.from_response(response, message=f"Failed to decode json body: {await response.text()}")
        except aiohttp.ClientPayloadError as e:
            raise MaelumaAPIError.from_response(
                response, message=f"An unexpected error occurred while receiving the response: {e}"
            )

        logger.debug(f"JSON response: {json_response}")
        self._check_response(json_response, response.headers, response.status)
        return json_response

    async def close(self):
        return await self._backend.close()

    async def __aenter__(self):
        if self._check_api_key_on_enter:
            await self.check_api_key()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    # API methods
    async def check_api_key(self) -> Dict[str, bool]:
        """
        check_api_key raises an exception when the key is invalid, but the return value for valid keys is kept for
        backwards compatibility.
        """
        return {"valid": is_api_key_valid(self.api_key)}

    async def batch_generate(
        self, prompts: List[str], return_exceptions=False, **kwargs
    ) -> List[Union[Exception, TextGenerations]]:
        return await asyncio.gather(
            *[self.generate(prompt, **kwargs) for prompt in prompts], return_exceptions=return_exceptions
        )

    async def generate(
        self,
        prompt: Optional[str] = None,
        prompt_vars: object = {},
        model: Optional[str] = None,
        preset: Optional[str] = None,
        num_generations: Optional[int] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        k: Optional[int] = None,
        p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        end_sequences: Optional[List[str]] = None,
        stop_sequences: Optional[List[str]] = None,
        return_likelihoods: Optional[str] = None,
        truncate: Optional[str] = None,
        logit_bias: Dict[int, float] = {},
        stream: bool = False,
    ) -> Union[TextGenerations, StreamingTextGenerations]:
        json_body = {
            "model": model,
            "prompt": prompt,
            "prompt_vars": prompt_vars,
            "preset": preset,
            "num_generations": num_generations,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "k": k,
            "p": p,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "end_sequences": end_sequences,
            "stop_sequences": stop_sequences,
            "return_likelihoods": return_likelihoods,
            "truncate": truncate,
            "logit_bias": logit_bias,
            "stream": stream,
        }
        response = await self._request(maeluma.GENERATE_URL, json=json_body, stream=stream)
        if stream:
            return StreamingTextGenerations(response)
        else:
            return TextGenerations.from_dict(response=response, return_likelihoods=return_likelihoods)

    async def chat(
        self,
        message: Optional[str] = None,
        query: Optional[str] = None,
        conversation_id: Optional[str] = "",
        model: Optional[str] = None,
        return_chatlog: Optional[bool] = False,
        return_prompt: Optional[bool] = False,
        return_preamble: Optional[bool] = False,
        chat_history: Optional[List[Dict[str, str]]] = None,
        preamble_override: Optional[str] = None,
        user_name: Optional[str] = None,
        temperature: Optional[float] = 0.8,
        max_tokens: Optional[int] = None,
        stream: Optional[bool] = False,
        p: Optional[float] = None,
        k: Optional[float] = None,
        logit_bias: Optional[Dict[int, float]] = None,
        mode: Optional[Mode] = None,
        documents: Optional[List[Dict[str, str]]] = None,
    ) -> Union[AsyncChat, StreamingChat]:
        if chat_history is not None:
            should_warn = True
            for entry in chat_history:
                if "text" in entry:
                    entry["message"] = entry["text"]

                if "text" in entry and should_warn:
                    logger.warning(
                        "The 'text' parameter is deprecated and will be removed in a future version of this function. "
                        + "Use 'message' instead.",
                    )
                    should_warn = False

        if query is None and message is None:
            raise MaelumaError("Either 'query' or 'message' must be provided.")

        if query is not None:
            logger.warning(
                "The 'query' parameter is deprecated and will be removed in a future version of this function. "
                + "Use 'message' instead.",
            )
            message = query

        json_body = {
            "message": message,
            "conversation_id": conversation_id,
            "model": model,
            "return_chatlog": return_chatlog,
            "return_prompt": return_prompt,
            "return_preamble": return_preamble,
            "chat_history": chat_history,
            "preamble_override": preamble_override,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": stream,
            "user_name": user_name,
            "p": p,
            "k": k,
            "logit_bias": logit_bias,
            "mode": mode,
            "documents": documents,
        }

        response = await self._request(maeluma.CHAT_URL, json=json_body, stream=stream)

        if stream:
            return StreamingChat(response)
        else:
            return AsyncChat.from_dict(response, message=message, client=self)

    async def embed(
        self,
        texts: List[str],
        model: Optional[str] = None,
        truncate: Optional[str] = None,
        compress: Optional[bool] = False,
        compression_codebook: Optional[str] = "default",
    ) -> Embeddings:
        """Returns an Embeddings object for the provided texts. Visit https://maeluma.ai/embed to learn about embeddings.

        Args:
            text (List[str]): A list of strings to embed.
            model (str): (Optional) The model ID to use for embedding the text.
            truncate (str): (Optional) One of NONE|START|END, defaults to END. How the API handles text longer than the maximum token length.
            compress (bool): (Optional) Whether to compress the embeddings. When True, the compressed_embeddings will be returned as integers in the range [0, 255].
            compression_codebook (str): (Optional) The compression codebook to use for compressed embeddings. Defaults to "default".
        """
        json_bodys = [
            dict(
                texts=texts[i : i + maeluma.MAELUMA_EMBED_BATCH_SIZE],
                model=model,
                truncate=truncate,
                compress=compress,
                compression_codebook=compression_codebook,
            )
            for i in range(0, len(texts), maeluma.MAELUMA_EMBED_BATCH_SIZE)
        ]
        responses = await asyncio.gather(*[self._request(maeluma.EMBED_URL, json) for json in json_bodys])
        meta = responses[0]["meta"] if responses else None

        return Embeddings(
            embeddings=[e for res in responses for e in res["embeddings"]],
            compressed_embeddings=[e for res in responses for e in res["compressed_embeddings"]] if compress else None,
            meta=meta,
        )


    async def batch_tokenize(
        self, texts: List[str], return_exceptions=False, **kwargs
    ) -> List[Union[Tokens, Exception]]:
        return await asyncio.gather(*[self.tokenize(t, **kwargs) for t in texts], return_exceptions=return_exceptions)

    async def tokenize(self, text: str, model: Optional[str] = None) -> Tokens:
        json_body = {"text": text, "model": model}
        res = await self._request(maeluma.TOKENIZE_URL, json_body)
        return Tokens(tokens=res["tokens"], token_strings=res["token_strings"], meta=res["meta"])

    async def batch_detokenize(
        self, list_of_tokens: List[List[int]], return_exceptions=False, **kwargs
    ) -> List[Union[Detokenization, Exception]]:
        return await asyncio.gather(
            *[self.detokenize(t, **kwargs) for t in list_of_tokens], return_exceptions=return_exceptions
        )

    async def detokenize(self, tokens: List[int], model: Optional[str] = None) -> Detokenization:
        json_body = {"tokens": tokens, "model": model}
        res = await self._request(maeluma.DETOKENIZE_URL, json_body)
        return Detokenization(text=res["text"], meta=res["meta"])

class AIOHTTPBackend:
    """HTTP backend which handles retries, concurrency limiting and logging"""

    SLEEP_AFTER_FAILURE = defaultdict(lambda: 0.25, {429: 5})

    def __init__(self, logger, max_concurrent_requests: int = 64, max_retries: int = 5, timeout: int = 120):
        self.logger = logger
        self.timeout = timeout
        self.max_retries = max_retries
        self.max_concurrent_requests = max_concurrent_requests
        self._semaphore: asyncio.Semaphore = None
        self._session: Optional[aiohttp.ClientSession] = None
        self._requester = None

    def build_aio_requester(self) -> Callable:  # returns a function for retryable requests
        @backoff.on_exception(
            backoff.expo,
            (aiohttp.ClientError, aiohttp.ClientResponseError),
            max_tries=self.max_retries + 1,
            max_time=self.timeout,
        )
        async def make_request_fn(session, *args, **kwargs):
            async with self._semaphore:  # this limits total concurrency by the client
                response = await session.request(*args, **kwargs)
            if response.status in maeluma.RETRY_STATUS_CODES:  # likely temporary, raise to retry
                self.logger.info(f"Received status {response.status}, retrying...")
                await asyncio.sleep(self.SLEEP_AFTER_FAILURE[response.status])
                response.raise_for_status()

            return response

        return make_request_fn

    async def request(
        self, url, json=None, method: str = "post", headers=None, session=None, stream=False, **kwargs
    ) -> JSON:
        session = session or await self.session()
        self.logger.debug(f"Making request to {url} with content {json}")

        request_start = time.time()
        try:
            response = await self._requester(session, method, url, headers=headers, json=json, **kwargs)
        except aiohttp.ClientConnectionError as e:  # ensure the SDK user does not have to deal with knowing aiohttp
            self.logger.debug(f"Fatal connection error after {time.time()-request_start:.1f}s: {e}")
            raise MaelumaConnectionError(str(e)) from e
        except aiohttp.ClientResponseError as e:  # status 500 or something remains after retries
            self.logger.debug(f"Fatal ClientResponseError error after {time.time()-request_start:.1f}s: {e}")
            raise MaelumaConnectionError(str(e)) from e
        except asyncio.TimeoutError as e:
            self.logger.debug(f"Fatal timeout error after {time.time()-request_start:.1f}s: {e}")
            raise MaelumaConnectionError("The request timed out") from e
        except Exception as e:  # Anything caught here should be added above
            self.logger.debug(f"Unexpected fatal error after {time.time()-request_start:.1f}s: {e}")
            raise MaelumaError(f"Unexpected exception ({e.__class__.__name__}): {e}") from e

        self.logger.debug(f"Received response with status {response.status} after {time.time()-request_start:.1f}s")
        return response

    async def session(self) -> aiohttp.ClientSession:
        if self._session is None:
            self._session = aiohttp.ClientSession(
                json_serialize=np_json_dumps,
                timeout=aiohttp.ClientTimeout(self.timeout),
                connector=aiohttp.TCPConnector(limit=0),
            )
            self._semaphore = asyncio.Semaphore(self.max_concurrent_requests)
            self._requester = self.build_aio_requester()
        return self._session

    async def close(self):
        if self._session is not None:
            await self._session.close()
            self._session = None

    def __del__(self):
        # https://stackoverflow.com/questions/54770360/how-can-i-wait-for-an-objects-del-to-finish-before-the-async-loop-closes
        if self._session:
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.create_task(self.close())
                else:
                    loop.run_until_complete(self.close())
            except Exception:
                pass


import abc
import asyncio
import json
import time
from concurrent import futures
from typing import Awaitable, Callable, Dict, List, Optional, TypeVar

from maeluma.error import MaelumaError

try:  # numpy is optional, but support json encoding if the user has it
    import numpy as np

    class MaelumaJsonEncoder(json.JSONEncoder):
        """Handles numpy datatypes and such in json encoding"""

        def default(self, obj):
            if isinstance(obj, np.integer):
                return int(obj)
            elif isinstance(obj, np.floating):
                return float(obj)
            elif isinstance(obj, np.ndarray):
                return obj.tolist()
            elif isinstance(obj, set):
                return list(obj)
            else:
                return super().default(obj)

except:

    class MaelumaJsonEncoder(json.JSONEncoder):
        """numpy is missing, so we can't handle these (and don't expect them)"""

        def default(self, obj):
            if isinstance(obj, set):
                return list(obj)
            else:
                return super().default(obj)


def np_json_dumps(data, **kwargs):
    return json.dumps(data, cls=MaelumaJsonEncoder, **kwargs)


def is_api_key_valid(key: Optional[str]) -> bool:
    """is_api_key_valid returns True when the key is valid and raises a MaelumaError when it is invalid."""
    if not key:
        raise MaelumaError(
            "No API key provided. Provide the API key in the client initialization or the CO_API_KEY environment variable."  # noqa: E501
        )

    return True


class JobWithStatus(abc.ABC):
    @abc.abstractmethod
    def has_terminal_status(self) -> bool:
        ...


T = TypeVar("T", bound=JobWithStatus)


def wait_for_job(
    get_job: Callable[[], T],
    timeout: Optional[float] = None,
    interval: float = 10,
) -> T:
    start_time = time.time()
    job = get_job()

    while not job.has_terminal_status():
        if timeout is not None and time.time() - start_time > timeout:
            raise TimeoutError(f"wait_for_job timed out after {timeout} seconds")

        time.sleep(interval)
        job = get_job()

    return job


async def async_wait_for_job(
    get_job: Callable[[], Awaitable[T]],
    timeout: Optional[float] = None,
    interval: float = 10,
) -> T:
    start_time = time.time()
    job = await get_job()

    while not job.has_terminal_status():
        if timeout is not None and time.time() - start_time > timeout:
            raise TimeoutError(f"wait_for_job timed out after {timeout} seconds")

        await asyncio.sleep(interval)
        job = await get_job()

    return job


def threadpool_map(f, call_args: List[Dict], num_workers, return_exceptions: bool = False) -> List:
    """Helper function similar to futures.ThreadPoolExecutor.map, but allows returning exceptions"""
    results = []
    with futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
        futures_list = [executor.submit(f, **args) for args in call_args]
        for future in futures_list:
            try:
                results.append(future.result())
            except Exception as e:
                if return_exceptions:
                    results.append(e)
                else:
                    raise
    return results


def get_api_key() -> str:
    return "api_key"


from importlib_metadata import version  # use package to support python 3.7

from maeluma.client import Client
from maeluma.client_async import AsyncClient
from maeluma.error import MaelumaAPIError, MaelumaConnectionError, MaelumaError

MAE_API_URL =  "http://127.0.0.1:8000"  #"https://api.maeluma.io"
RETRY_STATUS_CODES = [429, 500, 502, 503, 504]

API_VERSION = "1"
SDK_VERSION = version("maeluma")
MAELUMA_EMBED_BATCH_SIZE = 96
CHAT_URL = "chat"
CLASSIFY_URL = "classify"
CODEBOOK_URL = "embed-codebook"
DETECT_LANG_URL = "detect-language"
EMBED_URL = "embed"
GENERATE_FEEDBACK_URL = "feedback/generate"
GENERATE_PREFERENCE_FEEDBACK_URL = "feedback/generate/preference"
GENERATE_URL = "generate/text"
GENERATE_IMAGE_URL = "generate/image"
SUMMARIZE_URL = "summarize"
RERANK_URL = "rerank"
UPSCALE_IMAGE_URL = "upscale-image"

CHECK_API_KEY_URL = "check-api-key"
TOKENIZE_URL = "tokenize"
DETOKENIZE_URL = "detokenize"

CLUSTER_JOBS_URL = "cluster-jobs"
BULK_EMBED_JOBS_URL = "embed-jobs"
CUSTOM_MODEL_URL = "finetune"


import csv
from abc import ABC, abstractmethod
from io import StringIO
from pathlib import Path
from typing import Iterable, Optional, Tuple

try:
    from typing import TypedDict
except ImportError:
    from typing_extensions import TypedDict


class FileConfig(TypedDict):
    separator: str
    switchColumns: bool
    hasHeader: bool
    delimiter: str


_empty_file_config = FileConfig(separator="", switchColumns=False, hasHeader=False, delimiter="")


class CustomModelDataset(ABC):
    @abstractmethod
    def file_config(self) -> FileConfig:
        pass

    @abstractmethod
    def train_file_name(self) -> str:
        ...

    @abstractmethod
    def eval_file_name(self) -> str:
        ...

    @abstractmethod
    def has_eval_file(self) -> str:
        ...

    @abstractmethod
    def get_train_data(self) -> Iterable[bytes]:
        ...

    @abstractmethod
    def get_eval_data(self) -> Iterable[bytes]:
        ...


class LocalFileCustomModelDataset(CustomModelDataset):
    def __init__(self, train_file: str, eval_file: Optional[str] = None):
        self._train_file = train_file
        self._eval_file = eval_file

    @abstractmethod
    def file_config(self) -> FileConfig:
        pass

    def train_file_name(self) -> str:
        return Path(self._train_file).name

    def eval_file_name(self) -> str:
        if not self.has_eval_file():
            raise ValueError("Dataset has no eval file")
        return Path(self._eval_file).name

    def has_eval_file(self) -> str:
        return self._eval_file is not None

    def get_train_data(self) -> Iterable[bytes]:
        with open(self._train_file, mode="rb") as _f:
            for line in _f:
                yield line

    def get_eval_data(self) -> Iterable[bytes]:
        if not self.has_eval_file():
            raise ValueError("Dataset has no eval file")
        with open(self._eval_file, mode="rb") as _f:
            for line in _f:
                yield line


class CsvDataset(LocalFileCustomModelDataset):
    """
    A dataset consisting of local csv files. Each row should contain two items.
    E.g.: for prompt completion:
    this is the prompt,and this the completion
    another prompt, another completion
    """

    def __init__(self, train_file: str, delimiter: str, eval_file: Optional[str] = None):
        """
        Args:
            train_file (str): local path to csv with training data
            eval_file: (str, optional): local path to a csv with eval data
        """
        super().__init__(train_file, eval_file)
        self._delimiter = delimiter

    def file_config(self) -> FileConfig:
        config = dict(_empty_file_config)
        config["delimiter"] = self._delimiter
        return config


class JsonlDataset(LocalFileCustomModelDataset):
    """
    A dataset consisting of local jsonl files.
    Examples:
        prompt completion: {"prompt": "this is the prompt", "completion": "this is the completion"}
    """

    def file_config(self) -> FileConfig:
        return _empty_file_config.copy()


class TextDataset(LocalFileCustomModelDataset):
    """
    A dataset consisting of local text files. Can be used for generative custom models.
    """

    def __init__(self, train_file: str, separator: Optional[str], eval_file: Optional[Path] = None):
        super().__init__(train_file, eval_file)
        self._separator = separator

    def file_config(self) -> FileConfig:
        config = _empty_file_config.copy()
        if self._separator:
            config["separator"] = self._separator
        return config


class InMemoryDataset(CustomModelDataset):
    """
    A dataset existing in memory. You may pass a generator to avoid loading your whole dataset into memory at once.

    Examples:
        >>> InMemoryDataset([("this is a prompt", "this is a completion")])
        >>> InMemoryDataset([("example1", "label1"), ("example2", "label1"), ("example3", "label2")])
    """

    def __init__(self, training_data: Iterable[Tuple[str, str]], eval_data: Optional[Iterable[Tuple[str, str]]] = None):
        self._training_data = training_data
        self._eval_data = eval_data
        self._delimiter = ","

    def train_file_name(self) -> str:
        return "train.csv"

    def eval_file_name(self) -> str:
        if not self.has_eval_file():
            raise ValueError("Dataset has no eval data")
        return "eval.csv"

    def has_eval_file(self) -> str:
        return self._eval_data is not None

    def get_train_data(self) -> Iterable[bytes]:
        for row in self._training_data:
            yield self._serialize_row(row)

    def get_eval_data(self) -> Iterable[bytes]:
        if not self.has_eval_file():
            raise ValueError("Dataset has no eval data")
        for row in self._eval_data:
            yield self._serialize_row(row)

    def _serialize_row(self, row: Tuple[str, str]) -> bytes:
        buffer = StringIO()
        writer = csv.writer(buffer, delimiter=self._delimiter)
        writer.writerow(row)
        return buffer.getvalue().encode("utf-8")

    def file_config(self) -> FileConfig:
        config = _empty_file_config.copy()
        config["delimiter"] = self._delimiter
        return config






from typing import Any, Dict, Iterator, List, Optional
from dataclasses import dataclass, field

from maeluma.responses.base import MaelumaObject

@dataclass
class UpscaledImage(MaelumaObject):
    id: str
    url: str
    meta: Optional[dict]


from typing import Any, Dict, List, Optional

from maeluma.responses.base import MaelumaObject, _df_html


class Tokens(MaelumaObject):
    def __init__(
        self,
        tokens: Optional[List[int]] = None,
        token_strings: Optional[List[str]] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        assert tokens is not None
        assert token_strings is not None
        self.tokens = tokens
        self.token_strings = token_strings
        self.meta = meta

    @property
    def length(self):
        return len(self)

    def __len__(self) -> int:
        return len(self.tokens)

    def visualize(self, **kwargs):
        import pandas as pd

        df = pd.DataFrame.from_dict({"token_strings": self.token_strings, "tokens": self.tokens}, orient="index")
        return _df_html(df.fillna(""), style={"font-size": "90%"})


class Detokenization(MaelumaObject):
    def __init__(self, text: Optional[str] = None, meta: Optional[Dict[str, Any]] = None) -> None:
        assert text is not None
        self.text = text
        self.meta = meta

    def __str__(self) -> str:
        return self.text

    def __eq__(self, __o: object) -> bool:
        return self.text == __o.text


from .embeddings import Embeddings
from .text_generation import TextGeneration, StreamingTextGenerations, TextGenerations
from .image_generation import ImageGeneration, ImageGenerations
from .tokenize import Tokens, Detokenization
from .upscale import UpscaledImage
from .summarize import SummarizeResponse



from typing import Any, Dict, NamedTuple, Optional

SummarizeResponse = NamedTuple("SummarizeResponse", [("id", str), ("summary", str), ("meta", Optional[Dict[str, Any]])])
SummarizeResponse.__doc__ = """
Returned by ml.summarize, which generates a summary of the specified length for the provided text.

Example:
```
res = ml.summarize(text="Stock market report for today...")
print(res.summary)
```

Example:
```
res = ml.summarize(
    text="Stock market report for today...",
    model="summarize-xlarge",
    length="long",
    format="bullets",
    temperature=0.3,
    additional_command="focusing on the highest performing stocks")
print(res.summary)
```
"""


import json
from collections import UserList
from typing import Any, Dict, Generator, List, NamedTuple, Optional

import requests

from maeluma.responses.base import MaelumaObject, _df_html

TokenLikelihood = NamedTuple("TokenLikelihood", [("token", str), ("likelihood", float)])

TOKEN_COLORS = [
    (-2, "#FFECE2"),
    (-4, "#FFD6BC"),
    (-6, "#FFC59A"),
    (-8, "#FFB471"),
    (-10, "#FFA745"),
    (-12, "#FE9F00"),
    (-1e9, "#E18C00"),
]


class TextGeneration(MaelumaObject, str):
    def __new__(cls, text: str, *_, **__):
        return str.__new__(cls, text)

    def __init__(
        self,
        text: str,
        likelihood: float,
        token_likelihoods: List[TokenLikelihood] = [],
        prompt: str = None,
        finish_reason: str = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.prompt = prompt
        self.text = text
        self.likelihood = likelihood
        self.finish_reason = finish_reason
        self.token_likelihoods = token_likelihoods

    @classmethod
    def from_response(cls, response, prompt=None, **kwargs):
        token_likelihoods = response.get("token_likelihoods")
        if token_likelihoods:
            token_likelihoods = [TokenLikelihood(d["token"], d.get("likelihood")) for d in token_likelihoods]
        return cls(
            text=response.get("text"),
            likelihood=response.get("likelihood"),
            token_likelihoods=token_likelihoods,
            prompt=prompt,
            id=response.get("id"),
            finish_reason=response.get("finish_reason"),
            **kwargs,
        )

    # nice jupyter output
    def visualize_token_likelihoods(self, ignore_first_n=0, midpoint=-3, value_range=8, display=True):  # very WIP
        if self.token_likelihoods is None:
            return None

        def color_token(i, t: TokenLikelihood):
            if t.likelihood is None or i < ignore_first_n:
                col = "#EDEDED"
            else:
                col = next(c for thr, c in TOKEN_COLORS if t.likelihood >= thr)  # first hit
            return f"<span style='background-color:{col}'>{t.token}</span>"

        html = "".join(color_token(i, t) for i, t in enumerate(self.token_likelihoods))
        if display:
            from IPython.display import HTML

            return HTML(html)  # show in jupyter by default, but allow to be used as helper
        return html

    def _visualize_helper(self):
        return dict(
            prompt=self.prompt,
            text=self.text,
            likelihood=self.likelihood,
            token_likelihoods=self.visualize_token_likelihoods(display=False),
        )

    def visualize(self, **kwargs) -> str:
        import pandas as pd

        with pd.option_context("display.max_colwidth", 250):
            return _df_html(pd.DataFrame([self._visualize_helper()]), **kwargs)


class TextGenerations(UserList, MaelumaObject):
    def __init__(self, generations,  meta: Optional[Dict[str, Any]] = None) -> None:
        super().__init__(generations)
        self.meta = meta

    @classmethod
    def from_dict(cls, response: Dict[str, Any]) -> List[TextGeneration]:
        generations: List[TextGeneration] = []
        for gen in response["generations"]:
            likelihood = None
            token_likelihoods = None
            if "token_likelihoods" in gen.keys():
                token_likelihoods = []
                for likelihoods in gen["token_likelihoods"]:
                    token_likelihood = likelihoods["likelihood"] if "likelihood" in likelihoods.keys() else None
                    token_likelihoods.append(TokenLikelihood(likelihoods["token"], token_likelihood))
            generations.append(
                TextGeneration(
                    gen["text"],
                    likelihood,
                    token_likelihoods,
                    prompt=response.get("prompt"),
                    id=gen["id"],
                    finish_reason=gen.get("finish_reason"),
                )
            )

        return cls(generations, response.get("meta"))

    @property
    def generations(self) -> List[TextGeneration]:  # backward compatibility
        return self.data

    # nice jupyter output
    def visualize(self, **kwargs) -> str:
        import pandas as pd

        with pd.option_context("display.max_colwidth", 250):
            return _df_html(pd.DataFrame([g._visualize_helper() for g in self]), **kwargs)

    @property
    def prompt(self) -> str:
        """Returns the prompt used as input"""
        return self[0].prompt  # should all be the same


# ("likelihood", Optional[float])]) not supported
StreamingText = NamedTuple("StreamingText", [("index", Optional[int]), ("text", str), ("is_finished", bool)])


class StreamingTextGenerations(MaelumaObject):
    def __init__(self, response):
        self.response = response
        self.id = None
        self.generations = None
        self.finish_reason = None
        self.texts = []

    def _make_response_item(self, line) -> Optional[StreamingText]:
        streaming_item = json.loads(line)
        is_finished = streaming_item.get("is_finished")

        if not is_finished:
            index = streaming_item.get("index", 0)
            text = streaming_item.get("text")
            while len(self.texts) <= index:
                self.texts.append("")
            if text is None:
                return None
            self.texts[index] += text
            return StreamingText(text=text, is_finished=is_finished, index=index)

        self.finish_reason = streaming_item.get("finish_reason")
        generation_response = streaming_item.get("response")

        if generation_response is None:
            return None

        self.id = generation_response.get("id")
        # likelihoods not supported in streaming currently
        self.generations = TextGenerations.from_dict(generation_response)
        return None

    def __iter__(self) -> Generator[StreamingText, None, None]:
        if not isinstance(self.response, requests.Response):
            raise ValueError("For AsyncClient, use `async for` to iterate through the `StreamingGenerations`")

        for line in self.response.iter_lines():
            item = self._make_response_item(line)
            if item is not None:
                yield item

    async def __aiter__(self) -> Generator[StreamingText, None, None]:
        async for line in self.response.content:
            item = self._make_response_item(line)
            if item is not None:
                yield item